from .impl.adagrad import AdaGradClassifier
from .impl.primal_cd import CDClassifier
from .impl.fista import FistaClassifier
from .impl.sag import SAGClassifier
from .impl.sag import SAGAClassifier
from .impl.sdca import SDCAClassifier
from .impl.sgd import SGDClassifier
from .impl.svrg import SVRGClassifier
from .impl.dual_cd import LinearSVC
from .impl.primal_newton import KernelSVC
